/*
 * Asterisk -- A telephony toolkit for Linux.
 * Copyright (C) 1999-2005, Digium, Inc.
 * Mark Spencer <markster@digium.com>  - Asterisk main author
 *
 * app_dialipcontact.c: Dial IPContact applications
 *
 * Copyright (c) 2004-2008, IPContact Software SRL.  All rights reserved.
 * Authors:
 *
 * Luis Vazquez <luis@ipcontact.com.uy>
 * Daniel Ferrer <daniel@ipcontact.com.uy>
 *
 * $Id: app_dialipcontact.c,v 1.31 2007-11-19 15:05:45 ipcontact Exp $
 *
 * This code is released by the author with no restrictions on usage.
 *
 */


/*! \file
 *
 * \brief IPContact Dialing applications
 *
 * \author Luis Vazquez <luis@ipcontact.com.uy>
 * \author Daniel Ferrer <daniel@ipcontact.com.uy>
 * 
 * \ingroup applications
 */


/*** MODULEINFO
		 <depend>mysqlclient</depend>
		 <depend>res_ipcontact</depend>
***/

#include <asterisk.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <sys/socket.h>

#include <asterisk/logger.h>
#include <asterisk/options.h>
#include <asterisk/channel.h>
#include <asterisk/pbx.h>
#include <asterisk/utils.h>
#include <asterisk/callerid.h>
#include <asterisk/module.h>
#include <asterisk/config.h>
#include <asterisk/astdb.h>
#include <asterisk/indications.h>
#include <asterisk/manager.h>
#include <asterisk/say.h>
#include <asterisk/app.h>
#include <time.h>

//#include <pcre.h>


#include "res_ipcontact.h"


#define CALL_FORWARD_BASE "ipcontact/callforward"
#define CALL_FORWARD_ALL "all"
#define CALL_FORWARD_BUSY "busy"
#define CALL_FORWARD_UNAVAIL "unavail"

#define MODULE_DESC "IPContact Dialing Applications"

static char *app_dialhunt = "DialHuntGroup";
static char *synopsis_dialhunt = "Dial a number over given hunt group";
static char *descrip_dialhunt =
	"  DialHuntGroup(hunt" AST_STANDARD_APP_DELIM "number" AST_STANDARD_APP_DELIM "failonbusy" AST_STANDARD_APP_DELIM "dialopts):\n"
	"Find a free line and dial the passed number on given Hunt Group from MySQL DB.\n"
	"The failonbusy argument says if must give up when Dial returns BUSY condition.\n"
	"Returns whatever value the dial returns 0 or -1 when a line cannot be found.\n";
static int dialhunt_exec(struct ast_channel *chan, void *data);

static char *app_setcidext = "SetCIDFromExten";
static char *synopsis_setcidext = "Set CallerID as user's extension on SIP calls";
static char *descrip_setcidext =
	"SetCIDFromExten()\n"
	"  If the caller is a SIP channel reads the current value of cid number\n"
	"and if this is not a valid extension number uses the user name in the From\n"
	"of the SIP message as a key to search the extension in the IPContact DB.\n"
	"If found, this extension number is set as the new cid number.\n";
static int setcidext_exec(struct ast_channel *chan, void *data);

static char *app_callforward = "CallForward";
static char *synopsis_callforward = "Enable/Disable Call Forward on dialing extension";
static char *descrip_callforward =
	"CallForward(type[" AST_STANDARD_APP_DELIM "<on|off>])\n"
	"  Enable or disable call forwarding of given type (a:all, b:busy, u:unavail) on current extension.\n";
static int callforward_exec(struct ast_channel *chan, void *data);

static char *app_savecallinfo = "SaveClientCallInfo";
static char *synopsis_savecallinfo = "Writes client call info to MySQL database";
static char *descrip_savecallinfo =
	"SaveClientCallInfo()\n" "  Writes caller and called ID info into MySQL database.\n"
	"  This app generates events based on standard channel variables and also the following vars:\n "
    "INCOMING_CALL, OUTGOING_CALL, CALLEDNUMBER, CLIENT_DN, CLIENT_CN, AGENT_NAME, GROUP_ID.\n";
/* Applicacion SaveClientCallInfo
   Para determinar el destino de las llamadas debe
   estar definida la variable OUTGOING_CALL en llamadas salientes */
static int savecallinfo_exec(struct ast_channel *chan, void *data);

static char *app_vpbxgoto = "VirtualPBXGoto";
static char *synopsis_vpbxgoto = "VirtualPBXGoto([m|o|v])";
static char *descrip_vpbxgoto =
	"VirtualPBXGoto([m|o|v])\n"
	"  Go to Virtual PBX specific section:\n"
	"\t m => mainmenu, o => operator, v => voicemail.\n";
static int vpbxgoto_exec(struct ast_channel *chan, void *data);

static char *app_get_pin_profile = "GetPinProfile";
static char *synopsis_get_pin_profile = "GetPinProfile(pin)";
static char *descrip_get_pin_profile =
	"GetPinProfile(pin)\n"
	"  Get authentication for user associated with given pin number.\n"
	"Affected variables are: AUTH_user, AUTH_local, AUTH_longdist, AUTH_cellular, AUTH_inter.\n";
static int get_pin_profile_exec(struct ast_channel *chan, void *data);

static char *app_get_exten_profile = "GetExtenProfile";
static char *synopsis_get_exten_profile = "GetExtenProfile(exten)";
static char *descrip_get_exten_profile =
	"GetExtenProfile(exten)\n"
	"  Get authentication for an extension.\n"
	"Affected variables are: AUTH_local, AUTH_longdist, AUTH_cellular, AUTH_inter.\n";
static int get_exten_profile_exec(struct ast_channel *chan, void *data);


static const int first_digit_timeout = 5000;
static const int digit_timeout = 2000;
static char *indication_confirm_tone = "425/100,0/200,425/100,0/200,425/100,0/1000";

#define NO_CID_NAME ""

/* Globals */
static int debug = 0;

static const char config_file[] = "ipcontact/dialipcontact.conf";


static struct callinfo_config {
	int usedb;					/* Whether to save call info in SQL DB */
	int usemanager;				/* Whether to raise call info events to manager */
} callinfo = {
.usedb = 0,.usemanager = 1,};

static struct dialhunt_config {
	int failonbusy;				/* Whether to keep trying when dial returns BUSY */
	char dialopts[32];			/* Options to pass to Dial application */
} dialhunt = {
.failonbusy = 1,.dialopts = "",};


static struct ast_app *appDial = NULL;

#define MAX_HUNT_GROUPS 10
enum call_direction { INCOMING_CALL = 0, LOCAL_CALL, OUTGOING_CALL };

struct line_info {
	char number[8];
	char protocol[16];
	char prefix[32];
	char postfix[32];
};

struct dial_info {
	int vpbx;
	char exten[32];
	char protocol[16];
	char channel[32];
	char dialstring[64];
	char vm[16];
	int vmflag;
	int monitor;
};

struct exten_profile {
	char exten[32];
	unsigned char local;
	unsigned char national;
	unsigned char cellular;
	unsigned char inter;
};

/**** Utility functions ****/
inline static int indication_comfirm(struct ast_channel *chan);
inline static int playback_file(struct ast_channel *, const char *);
inline static int build_dial_string(struct line_info *line, const char *number, char *dialstr, const int len);
/* fix_sip_cidnumber:
   Recibe un canal SIP y en caso de que el cid number no sea una extension valida
   busca si el canal sip esta asociado a alguna extension en la base de datos.
   En caso de encontrarla setea este valor como nuevo cid number y crea el nuevo
   string del callerID en el buffer que se pasa como parmetro.
   Devuelve 0 si se pudo modificar el cid number por una extension y -1 en otro caso
   El parametro cidnumber se carga con el valor de cid number en caso de exito en
   encontrar dicho valor */
static int fix_sip_cidnumber(struct ast_channel *chan);


/* astdb_get_exten_data(const char *exten, const char *key, char *value, int len);
   obtiene de la base de datos de asterisk el valor correspondiente
   el string key asociado a la extension exten
*/
static int astdb_get_exten_data(const char *exten, const char *key, char *value, int len);


static int get_exten_name(const char *exten, char *value, int len);
static int get_exten_dial_info(char *exten, struct dial_info *info);

/**** Virtual PBX functions ****/
inline static int find_current_vpbx(struct ast_channel *chan);
static int get_exten_vpbx(char *exten);
inline void show_vpbxs(void);


inline static int indication_comfirm(struct ast_channel *chan)
{
	int res;
	ast_safe_sleep(chan, 500);
	res = ast_playtones_start(chan, 0, indication_confirm_tone, 0);
	ast_safe_sleep(chan, 3000);
	ast_playtones_stop(chan);
	return res;
};

inline static int playback_file(struct ast_channel *chan, const char *filename)
{
	int res = 0;
	if (chan->_state != AST_STATE_UP) {
		/* Otherwise answer the line */
		res = ast_answer(chan);
	}
	if (!res) {
		ast_stopstream(chan);
		res = ast_streamfile(chan, (char *) filename, chan->language);
		if (res == 0)
			res = ast_waitstream(chan, "");
		else {
			ast_log(LOG_WARNING, "ast_streamfile failed on %s for %s\n", chan->name,
					(char *) filename);
			res = 0;
		}
		ast_stopstream(chan);
	}
	return res;
}

inline static int find_current_vpbx(struct ast_channel *chan)
{

	const char *vpbx_str = NULL;
	int vpbx = 0;

	/* Find the virtual pbx this call belongs to */
	vpbx_str = pbx_builtin_getvar_helper(chan, "VPBX");
	if (vpbx_str) {
		if (sscanf(vpbx_str, "%d", &vpbx)) {
			if (debug) {
				ast_log(LOG_NOTICE, "Call on virtual pbx %d\n", vpbx);
			}
		} else {
			ast_log(LOG_ERROR, "VPBX variable (%s) must be numeric\n", vpbx_str);
		}
	} else {
		ast_log(LOG_WARNING, "VPBX not set, using 0 as default value\n");
		vpbx = 0;				/* default value */
	}

	return vpbx;
}

static int astdb_get_exten_data(const char *exten, const char *key, char *value, int len)
{
	int res = 1;
	char family[256] = "";

	if (!value) {
		return -1;
	}
	if (ast_strlen_zero(exten)) {
		return 0;
	}
	if (ast_strlen_zero(key)) {
		return 0;
	}
	snprintf(family, sizeof(family), "%s/%s", IPC_EXTEN_TREE, exten);
	if (!ast_db_get(family, key, value, len)) {
		if (option_verbose > 2) {
			ast_verbose(VERBOSE_PREFIX_3 "AstDB value is %s for key %s at extension %s\n",
						value, key, exten);
		}
		res = 1;
	} else {
		ast_log(LOG_NOTICE, "AstDB has no value for key %s at extension %s\n", key,
				exten);
		value[0]=0;
		res = 0;
	}
	return res;
}


static int get_exten_name(const char *exten, char *buffer, int len)
{
	int res;

	if (!buffer)
		return 0;

	if (!is_number(exten)) {
		ast_log(LOG_DEBUG, "Trying to get name for non numeric extension: %s\n", exten);
		return 0;
	}
	res = astdb_get_exten_data(exten, IPC_EXTEN_INFO, buffer, len);
	if (res && !ast_strlen_zero(buffer)) {
		if (debug)
			ast_log(LOG_NOTICE, "Found name %s for %s.\n", buffer, exten);
		res = 1;
	} else {
		res = 0;
		ast_log(LOG_NOTICE, "No Name found for %s.\n", exten);
	}
	return res;
}

static int get_exten_vpbx(char *exten)
{
	char dbres[32] = "";
	int vpbx = 0;				// Default value
	int res = 0;
	res = astdb_get_exten_data(exten, IPC_EXTEN_NUMBER, dbres, sizeof(dbres));
	if (res && !ast_strlen_zero(dbres)) {
		vpbx = strtol(dbres, (char **) NULL, 10);
	} else {
		ast_log(LOG_NOTICE, "No virtual pbx for extension %s, using 0\n", exten);
		vpbx = 0;
	}
	return vpbx;
}


static int get_exten_dial_info(char *exten, struct dial_info *info)
{
	int res = 0;
	char dbres[64] = "";
	if (!info)
		return 0;

	memset(info, 0, sizeof(struct dial_info));

	res = astdb_get_exten_data(exten, IPC_EXTEN_PROTO, dbres, sizeof(dbres));
	if (res && !ast_strlen_zero(dbres)) {	// Found protocol
		ast_copy_string(info->protocol, dbres, sizeof(info->protocol));
		ast_copy_string(info->exten, exten, sizeof(info->exten));
		// Virtual PBX for this extension
		info->vpbx = get_exten_vpbx(exten);

		// Voicemail info
		res = astdb_get_exten_data(exten, IPC_EXTEN_VMFLAG, dbres, sizeof(dbres));
		if (res && ast_true(dbres)) {
			res = astdb_get_exten_data(exten, IPC_EXTEN_VM, dbres, sizeof(dbres));
			if (res && !ast_strlen_zero(dbres)) {
				ast_copy_string(info->vm, dbres, sizeof(info->vm));
				info->vmflag = 1;
			} else {
				info->vmflag = 0;
			}
		} else {
			info->vmflag = 0;
		}

		// Monitor(record) this extension
		res = astdb_get_exten_data(exten, IPC_EXTEN_MONITOR, dbres, sizeof(dbres));
		if (res && ast_true(dbres)) {
			info->monitor = 1;
		}
		// Channel to dial for this extension
		res = astdb_get_exten_data(exten, IPC_EXTEN_CHANNEL, dbres, sizeof(dbres));
		if (res && !ast_strlen_zero(dbres)) {	// Found channel
			ast_copy_string(info->channel, dbres, sizeof(info->channel));
			if (!strcasecmp(info->protocol, "sip")) {
				snprintf(info->dialstring, sizeof(info->dialstring), "SIP/%s",
						 info->channel);
			} else if (!strcasecmp(info->protocol, "mgcp")) {
				snprintf(info->dialstring, sizeof(info->dialstring), "MGCP/aaln/%s",
						 info->channel);
			} else if (!strcasecmp(info->protocol, "zap")) {
				snprintf(info->dialstring, sizeof(info->dialstring), "ZAP/%s",
						 info->channel);
			} else if (!strcasecmp(info->protocol, "h323")) {
				snprintf(info->dialstring, sizeof(info->dialstring), "H323/%s",
						 info->channel);
			} else if (!strcasecmp(info->protocol, "oh323")) {
				snprintf(info->dialstring, sizeof(info->dialstring), "OH323/%s",
						 info->channel);
			} else {
				ast_log(LOG_ERROR, "Unknown protocol: %s\n", info->protocol);
				res = -2;
			}
		} else {
			ast_log(LOG_WARNING,
					"Unable to create dial string for exten %s: empty channel field\n",
					exten);
			res = -1;
		}
	} else {
		ast_log(LOG_WARNING,
				"Unable to create dial string for exten %s: empty protocol field\n",
				exten);
		res = -1;
	}

	if (res == 0 && debug)
		ast_log(LOG_NOTICE, "Dial string result: %s\n", info->dialstring);
	return res;
}



// Applicacion SaveClientCallInfo
static int savecallinfo_exec(struct ast_channel *chan, void *data)
{
	MYSQL *conn = NULL;
	char *cidnumber = "", *cidname = "";
	char name_buffer[128], dnid[32];
	const char *clientid = NULL, *called_num = NULL, *dnidname = NULL, *groupid = NULL, *agent = NULL;
	char *direction;
	char sqlbuffer[IPC_QUERY_LEN];
	enum call_direction call_dir;
	int is_update;
	struct ast_module_user *u;

	u = ast_module_user_add(chan);

	memset(sqlbuffer, 0, sizeof(sqlbuffer));

	if (!ast_strlen_zero(chan->cid.cid_num)) {
		cidnumber = chan->cid.cid_num;
	}
	if (!ast_strlen_zero(chan->cid.cid_name)) {
		cidname = chan->cid.cid_name;
	}
	if (debug) {
		ast_log(LOG_NOTICE, "CallerID parsed as number='%s', name='%s'\n", cidnumber,
				cidname);
	}
	if (ast_true(pbx_builtin_getvar_helper(chan, "OUTGOING_CALL"))) {
		call_dir = OUTGOING_CALL;
	} else if (IS_INCOMING_CALL(chan)) {
		call_dir = INCOMING_CALL;
	} else {
		call_dir = LOCAL_CALL;
	}

 	agent = pbx_builtin_getvar_helper(chan, "AGENT_NAME");
	if (agent && debug)
		ast_log(LOG_DEBUG, "Using Agent Name: %s", agent);
	groupid = pbx_builtin_getvar_helper(chan, "GROUP_ID");
	if (groupid && debug)
		ast_log(LOG_DEBUG, "Using GroupID: %s", groupid);

	switch (call_dir) {
	case INCOMING_CALL:		// Incoming call
		direction = "in";
		clientid = pbx_builtin_getvar_helper(chan, "CLIENT_DN");
		/******** Destination ID **********/
		called_num = pbx_builtin_getvar_helper(chan, "CALLEDNUMBER");
		if (!ast_strlen_zero(called_num)) {
			ast_copy_string(dnid, called_num, sizeof(dnid));
		} else if (!ast_strlen_zero(chan->cid.cid_dnid)) {
			ast_copy_string(dnid, chan->cid.cid_dnid, sizeof(dnid));
		} else {
			ast_log(LOG_DEBUG,
					"No DNID found in channel. Using current extension instead.\n");
			ast_copy_string(dnid, CHAN_REAL_EXTEN(chan), sizeof(dnid));
		};
		if (get_exten_name(CHAN_REAL_EXTEN(chan), name_buffer, sizeof(name_buffer))) {
			dnidname = ast_strdupa(name_buffer);
		} else {
			dnidname = "";
		}
		break;
	case OUTGOING_CALL:
		// Outgoing call
		direction = "out";
		clientid = pbx_builtin_getvar_helper(chan, "CLIENT_DN");
		/********** Caller ID ************/
		if (get_exten_name(cidnumber, name_buffer, sizeof(name_buffer))) {
			cidname = ast_strdupa(name_buffer);
		} else {
			cidname = "";
		}
		/******** Destination ID **********/
		ast_copy_string(dnid, CHAN_REAL_EXTEN(chan), sizeof(dnid));
		dnidname = pbx_builtin_getvar_helper(chan, "CLIENT_CN");
		if (dnidname == NULL) {
			dnidname = "";
		}
		break;
	default:
		// Local call
		direction = "lo";
		/********** Caller ID	 ************/
		if (get_exten_name(cidnumber, name_buffer, sizeof(name_buffer))) {
			cidname = ast_strdupa(name_buffer);
			if (debug)
				ast_log(LOG_NOTICE, "Found name %s for %s.\n", cidname, cidnumber);
		} else {
			cidname = "";
		}
		/******** Destination ID **********/
		ast_copy_string(dnid, CHAN_REAL_EXTEN(chan), sizeof(dnid));
		dnidname = "";
		if (get_exten_name(dnid, name_buffer, sizeof(name_buffer))) {
			dnidname = ast_strdupa(name_buffer);
			if (debug)
				ast_log(LOG_DEBUG, "Found name %s for %s.\n", dnidname, dnid);
		}
		ast_copy_string(dnid, CHAN_REAL_EXTEN(chan), sizeof(dnid));
	}
	is_update = ast_true(pbx_builtin_getvar_helper(chan, "SAVED_CALL"));
	if (callinfo.usemanager) {
		if (is_update) {
			manager_event(EVENT_FLAG_AGENT, "UpdateClientCall",
						  "Uniqueid: %s\r\n"
						  "CallerID: %s\r\n"
						  "CallerIDName: %s\r\n"
						  "DNID: %s\r\n"
						  "DNIDName: %s\r\n"
						  "AgentName: %s\r\n"
						  "GroupID: %s\r\n",
						  chan->uniqueid, cidnumber, cidname, dnid, dnidname, agent ? agent : "", groupid ? groupid : "");
		} else {
			manager_event(EVENT_FLAG_AGENT, "SaveClientCall",
						  "Direction: %s\r\n"
						  "Uniqueid: %s\r\n"
						  "CallerID: %s\r\n"
						  "CallerIDName: %s\r\n"
						  "DNID: %s\r\n"
						  "DNIDName: %s\r\n"
						  "ClientID: %s\r\n",
						  direction, chan->uniqueid,
						  cidnumber, cidname, dnid, dnidname, clientid ? clientid : "");
		}
	}
	if (callinfo.usedb) {
	/***** BEGIN MUTEX BLOCK *******/
		ipc_mutex_lock();
		conn = ipc_sql_reconnect();
		if (conn) {
			char *end = sqlbuffer;
			if (is_update) {
				end = stpcpy(sqlbuffer, "UPDATE callinfo SET cid=");
				*end++ = '\'';
				end += mysql_real_escape_string(conn, end, cidnumber, strlen(cidnumber));
				*end++ = '\'';
				end = stpcpy(end, ",cidname=");
				*end++ = '\'';
				end += mysql_real_escape_string(conn, end, cidname, strlen(cidname));
				*end++ = '\'';
				end = stpcpy(end, ",dnid=");
				*end++ = '\'';
				end += mysql_real_escape_string(conn, end, dnid, strlen(dnid));
				*end++ = '\'';
				end = stpcpy(end, ",dnidname=");
				*end++ = '\'';
				end += mysql_real_escape_string(conn, end, dnidname, strlen(dnidname));
				*end++ = '\'';
				end = stpcpy(end, " WHERE uniqueid=");
				*end++ = '\'';
				end +=
					mysql_real_escape_string(conn, end, chan->uniqueid,
											 strlen(chan->uniqueid));
				*end++ = '\'';
			} else {
				end = stpcpy(sqlbuffer,
							 "INSERT INTO callinfo (uniqueid,direction,cid,cidname,dnid,dnidname,clientid) values (");
				*end++ = '\'';
				end += mysql_real_escape_string(conn, end, chan->uniqueid, strlen(chan->uniqueid));
				*end++ = '\'';
				*end++ = ',';
				*end++ = '\'';
				end += mysql_real_escape_string(conn, end, direction, strlen(direction));
				*end++ = '\'';
				*end++ = ',';
				*end++ = '\'';
				end += mysql_real_escape_string(conn, end, cidnumber, strlen(cidnumber));
				*end++ = '\'';
				*end++ = ',';
				*end++ = '\'';
				end += mysql_real_escape_string(conn, end, cidname, strlen(cidname));
				*end++ = '\'';
				*end++ = ',';
				*end++ = '\'';
				end += mysql_real_escape_string(conn, end, dnid, strlen(dnid));
				*end++ = '\'';
				*end++ = ',';
				*end++ = '\'';
				end += mysql_real_escape_string(conn, end, dnidname, strlen(dnidname));
				*end++ = '\'';
				*end++ = ',';
				*end++ = '\'';
				if (clientid)
					end +=
						mysql_real_escape_string(conn, end, clientid, strlen(clientid));
				*end++ = '\'';
				*end++ = ')';
			}
			*end = '\0';
			if (mysql_real_query(conn, sqlbuffer, (size_t)(end - sqlbuffer))) {
				ast_log(LOG_ERROR, "Database query failed: %s.\n", sqlbuffer);
			}
		} else {
			ast_log(LOG_ERROR, "Unable to connect to MySQL server\n");
		};
		/***** END MUTEX BLOCK *******/
		ipc_mutex_unlock();
	}
	pbx_builtin_setvar_helper(chan, "SAVED_CALL", "1");
	ast_module_user_remove(u);
	return 0;
}


static int fix_sip_cidnumber(struct ast_channel *chan)
{
	MYSQL_RES *sqlres = NULL;
	MYSQL_ROW row;
	char sql[IPC_QUERY_LEN] = "";
	int res = 0;

	if (!chan->cid.cid_num) {	/* No cid number found in channel */
		ast_log(LOG_NOTICE, "No Caller ID info in channel %s\n", chan->name);
		return -1;
	}

	snprintf(sql, (int) sizeof(sql),
			 "SELECT number FROM extension WHERE protocol LIKE 'SIP' AND channel LIKE '%s@%%'",
			 chan->cid.cid_num);

	sqlres = ipc_sql_exec_query(sql);

	if ((sqlres != NULL) && (mysql_num_rows(sqlres) > 0)) {
		row = mysql_fetch_row(sqlres);
		if ((row != NULL) && !ast_strlen_zero(row[0])) {
			if (debug) {
				ast_log(LOG_NOTICE, "Found extension = %s for user %s\n", row[0],
						chan->cid.cid_num);
			};
			ast_set_callerid(chan, row[0],
							 ast_strlen_zero(chan->cid.cid_name) ? NO_CID_NAME : NULL,
							 NULL);
		} else {
			res = -1;
		};
		mysql_free_result(sqlres);
	} else if (debug) {
		ast_log(LOG_NOTICE, "No extension found for SIP user %s\n", chan->cid.cid_num);
		res = -1;
	};
	return res;
}

static int setcidext_exec(struct ast_channel *chan, void *data)
{
	if (!strcmp(ast_channel_type(chan), "SIP")) {
		fix_sip_cidnumber(chan);
	};
	return 0;
}

static int callforward_exec(struct ast_channel *chan, void *data)
{
	int res = 0;
	int enable = 0;
	char newexten[32];
	char family[IPC_MAX_LEN];
	const char *fwd_type = NULL;
	int usage = 0;
	char *parse = NULL;
	struct ast_module_user *u;

	AST_DECLARE_APP_ARGS(args, AST_APP_ARG(type);
						 AST_APP_ARG(mode);
		);

	if (ast_strlen_zero(data)) {
		usage = 1;
	}
	parse = ast_strdupa(data);
	if (!parse) {
		ast_log(LOG_ERROR, "Unable to dupe data!");
		return -1;
	}

	AST_STANDARD_APP_ARGS(args, parse);

	if (ast_strlen_zero(args.type)) {
		usage = 1;
	} else {
		if (strncasecmp(args.type, "a", 1))
			fwd_type = CALL_FORWARD_ALL;
		else if (strncasecmp(args.type, "b", 1))
			fwd_type = CALL_FORWARD_BUSY;
		else if (strncasecmp(args.type, "u", 1))
			fwd_type = CALL_FORWARD_UNAVAIL;
		else
			usage = 1;
	};

	if (usage) {
		ast_log(LOG_ERROR,
				"Call Forward requires a parameter: type[" AST_STANDARD_APP_DELIM "<on|off>]. type: (a|u|b).\n");
		return -1;
	}

	u = ast_module_user_add(chan);

	if (ast_strlen_zero(args.mode))
		enable = 1;
	else
		enable = ast_true(args.mode);
	memset(newexten, 0, sizeof(newexten));

	if (!strcmp(ast_channel_type(chan), "SIP"))
		res = fix_sip_cidnumber(chan);

	if (ast_strlen_zero(chan->cid.cid_num)) {
		ast_log(LOG_ERROR, "Found empty CID number, unable to exec.\n");
		res = -1;
	};
	/* Answer if need be */
	if (chan->_state != AST_STATE_UP) {
		res = ast_answer(chan);
	}
	snprintf(family, sizeof(family), "%s/%s", CALL_FORWARD_BASE, fwd_type);
	if (!res) {
		/* Stop anything playing */
		ast_stopstream(chan);

		if (enable) {			/* Enable call forwarding in this extension */
			ast_safe_sleep(chan, 500);
			if ((res = ast_streamfile(chan, "beep", chan->language))) {
				ast_log(LOG_WARNING, "Unable to stream beep file.\n");
			} else {
				res = ast_waitstream(chan, "");
			}
			ast_stopstream(chan);

			res =
				ast_readstring(chan, newexten, sizeof(newexten) - 1, digit_timeout,
							   first_digit_timeout, "#");

			if ((res < 0) ||
				!ast_exists_extension(chan, chan->context, newexten, 1,
									  chan->cid.cid_num)) {
				/* Error or received no valid extension */
				if (debug) {
					if (!ast_strlen_zero(newexten))
						ast_log(LOG_WARNING, "Invalid extension: %s.\n", newexten);
					else
						ast_log(LOG_WARNING, "No digits received!!");
				}
				res = ast_streamfile(chan, "pbx-invalid", chan->language);
				if (!res) {
					res = ast_waitstream(chan, "");
				}
				ast_stopstream(chan);
				goto end;
			} else {			/* We got a valid extension */
				if (ast_db_put(family, chan->cid.cid_num, newexten)) {
					ast_log(LOG_WARNING,
							"Error writing value to database entry: %s/%s.\n", family,
							chan->cid.cid_num);
					res = -1;
				} else {
					ast_say_number(chan, atoi((char *) newexten), "", chan->language,
								   NULL);
					res = indication_comfirm(chan);
				}
			};
		} else {				/* Disable call forwarding in this extension */
			if (ast_db_del(family, chan->cid.cid_num)) {
				ast_log(LOG_WARNING, "Error deleting database entry: %s/%s.\n",
						family, chan->cid.cid_num);
				res = -1;
			} else {
				res = indication_comfirm(chan);
			};
		};
	} else {
		ast_log(LOG_WARNING, "Unable to answer channel %s.\n", chan->name);
	};

  end:
	ast_module_user_remove(u);
	return res;
}

static int get_pin_profile_exec(struct ast_channel *chan, void *data)
{
	int res = 0;
	MYSQL_RES *sqlres = NULL;
	MYSQL_ROW rowval;
	MYSQL_FIELD *fields = NULL;
	char query[IPC_QUERY_LEN];
	int numfields;
	int pin = 0;
	struct ast_module_user *u;

	if (ast_strlen_zero(data) || (pin = atoi(data)) < 0) {
		ast_log(LOG_WARNING, "GetPinProfile requires a numeric argument: pin\n");
		return -1;
	}

	u = ast_module_user_add(chan);

	/* Busco los permisos asociados a este pin */
	snprintf(query, sizeof(query),
			 "SELECT username,local,longdist,cellular,inter FROM authcalls WHERE pin=%d AND enabled",
			 pin);
	sqlres = ipc_sql_exec_query(query);
	if (sqlres && (mysql_num_rows(sqlres) > 0)) {
		/* Hubo resultado en la consulta */
		numfields = mysql_num_fields(sqlres);
		fields = mysql_fetch_fields(sqlres);
		/* Obtengo la primer fila */
		rowval = mysql_fetch_row(sqlres);
		int i;
		for (i = 0; i < numfields; i++) {
			if (rowval[i]) {
				if (!strcmp(fields[i].name, "username")) {
					// username is not defined
					if (ast_strlen_zero(rowval[i])) {
						pbx_builtin_setvar_helper(chan, "AUTH_user", "");
						break;
					} else {
						pbx_builtin_setvar_helper(chan, "AUTH_user", rowval[i]);
					}
				} else if (!strcmp(fields[i].name, "local")) {
					pbx_builtin_setvar_helper(chan, "AUTH_local", rowval[i]);
				} else if (!strcmp(fields[i].name, "longdist")) {
					pbx_builtin_setvar_helper(chan, "AUTH_longdist", rowval[i]);
				} else if (!strcmp(fields[i].name, "cellular")) {
					pbx_builtin_setvar_helper(chan, "AUTH_cellular", rowval[i]);
				} else if (!strcmp(fields[i].name, "inter")) {
					pbx_builtin_setvar_helper(chan, "AUTH_inter", rowval[i]);
				}
			}
		}
	} else {
		pbx_builtin_setvar_helper(chan, "AUTH_user", "");
	}
	ast_module_user_remove(u);
	return res;
}


static int get_exten_profile_exec(struct ast_channel *chan, void *data)
{
	int res = 0;
	MYSQL_RES *sqlres = NULL;
	MYSQL_ROW rowval;
	MYSQL_FIELD *fields = NULL;
	char query[IPC_QUERY_LEN];
	int numfields;
	int pin = 0;
	struct ast_module_user *u;

	if (ast_strlen_zero(data) || (pin = atoi(data)) < 0) {
		ast_log(LOG_WARNING, "GetExtenProfile requires an extension argument\n");
		return -1;
	}

	u = ast_module_user_add(chan);

	/* Busco los permisos asociados a este pin */
	snprintf(query, sizeof(query),
			 "SELECT local,longdist,cellular,inter FROM extension WHERE number='%s'",
			 (char*)data);
	sqlres = ipc_sql_exec_query(query);
	if (sqlres && (mysql_num_rows(sqlres) > 0)) {
		/* Hubo resultado en la consulta */
		numfields = mysql_num_fields(sqlres);
		fields = mysql_fetch_fields(sqlres);
		/* Obtengo la primer fila */
		rowval = mysql_fetch_row(sqlres);
		int i;
		for (i = 0; i < numfields; i++) {
			if (rowval[i]) {
				if (!strcmp(fields[i].name, "local")) {
					pbx_builtin_setvar_helper(chan, "AUTH_local", rowval[i]);
				} else if (!strcmp(fields[i].name, "longdist")) {
					pbx_builtin_setvar_helper(chan, "AUTH_longdist", rowval[i]);
				} else if (!strcmp(fields[i].name, "cellular")) {
					pbx_builtin_setvar_helper(chan, "AUTH_cellular", rowval[i]);
				} else if (!strcmp(fields[i].name, "inter")) {
					pbx_builtin_setvar_helper(chan, "AUTH_inter", rowval[i]);
				}
			}
		}
	}
	ast_module_user_remove(u);
	return res;
}


static int dialhunt_exec(struct ast_channel *chan, void *data)
{
	int res = 0;
	MYSQL_RES *sqlres = NULL;
	MYSQL_ROW rowval;
	MYSQL_FIELD *fields = NULL;
	int numfields;
	char dialargs[IPC_MAX_LEN];
	char query[IPC_QUERY_LEN];
	char var_buffer[32];
	int vpbx = 0;
	int huntgroup = 0;
	int hg = 1;
	long unsigned numlines = 0;
	struct line_info *line;
	int ready = 0;
	const char *dialstatus = "INVALIDARGS";
	int failonbusy = 1;
	struct ast_frame *frm;
	time_t rstart;
	struct ast_module_user *u;
	char *parse = NULL;
	AST_DECLARE_APP_ARGS(args,
						 AST_APP_ARG(huntgroup);
						 AST_APP_ARG(number); AST_APP_ARG(failonbusy);
						 AST_APP_ARG(dialopts);
		);

	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING,
				"DialHuntGroup requires an argument (hunt" AST_STANDARD_APP_DELIM "number" AST_STANDARD_APP_DELIM "failonbusy" AST_STANDARD_APP_DELIM "dialopts)\n");
		pbx_builtin_setvar_helper(chan, "DIALSTATUS", dialstatus);
		return -1;
	}

	parse = ast_strdupa(data);
	if (!parse) {
		ast_log(LOG_ERROR, "Unable to dupe data :(\n");
		return -1;
	}
	AST_STANDARD_APP_ARGS(args, parse);

	if (ast_strlen_zero(args.huntgroup) || ((huntgroup = atoi(args.huntgroup)) < 0)) {
		ast_log(LOG_ERROR, "DialHuntGroup requires a numeric huntgroup\n");
		pbx_builtin_setvar_helper(chan, "DIALSTATUS", dialstatus);
		return -1;
	}

	if (ast_strlen_zero(args.number)) {
		ast_log(LOG_ERROR, "DialHuntGroup requires a number to dial\n");
		pbx_builtin_setvar_helper(chan, "DIALSTATUS", dialstatus);
		return -1;
	}
	// Everything ok, let's increment module usage
	u = ast_module_user_add(chan);

	line = malloc(sizeof(struct line_info));
	memset(var_buffer, 0, sizeof(var_buffer));

	if (ast_strlen_zero(args.failonbusy)) {
		ast_log(LOG_WARNING,
				"DialHuntGroup called without failonbusy option. Defaults to yes.\n");
		failonbusy = dialhunt.failonbusy;	// Defaults to app config
	} else {
		failonbusy = ast_true(args.failonbusy);
	}

	if (ast_strlen_zero(args.dialopts)) {
		if (ast_strlen_zero(dialhunt.dialopts)) {
			args.dialopts = "";
			ast_log(LOG_NOTICE, "Dialing without dial options.\n");
		} else {
			args.dialopts = dialhunt.dialopts;
			if (debug)
				ast_log(LOG_NOTICE,
						"No dial options. Taking defaults from app config.\n");
		}
	}

	if (ast_strlen_zero(chan->cid.cid_num)) {
		ast_log(LOG_WARNING,
				"Unable to get valid extension from cid, using virtual pbx 0");
	} else {
		vpbx = get_exten_vpbx(chan->cid.cid_num);
	}

	snprintf(var_buffer, sizeof(var_buffer), "%d", vpbx);
	pbx_builtin_setvar_helper(chan, "VPBX", var_buffer);

	/* Busco las lineas en este huntgroup */
	snprintf(query, sizeof(query),
			 "SELECT number,protocol,line_prefix,line_postfix FROM pstnline as l, hunt_line as h WHERE l.number = h.line_number AND h.hunt_number = %d  AND enabled = 1 AND busy = 0 ORDER BY id",
			 huntgroup);
	sqlres = ipc_sql_exec_query(query);
	if (sqlres) {
		numlines = mysql_num_rows(sqlres);
	} else {
		res = -1;
	}

	if ((res == 0) && (numlines > 0)) {
		/* Hubo resultado en la consulta */
		/* y hay lineas libres en este huntgroup */
		snprintf(var_buffer, sizeof(var_buffer), "%d", hg);
		pbx_builtin_setvar_helper(chan, "HUNTGROUP", var_buffer);

		if (debug) {
			ast_log(LOG_NOTICE, "Have %lu free lines on huntgroup %d \n", numlines, hg);
		}
		// Itero sobre las lineas encontradas en este HUNTGROUP
		int i;
		while ((rowval = mysql_fetch_row(sqlres))) {
			memset(line, 0, sizeof(struct line_info));
			numfields = mysql_num_fields(sqlres);
			fields = mysql_fetch_fields(sqlres);
			for (i = 0; i < numfields; i++) {
				if (rowval[i]) {
					if (!strcmp(fields[i].name, "number")) {
						ast_copy_string(line->number, rowval[i], sizeof(line->number));
					} else if (!strcmp(fields[i].name, "protocol")) {
						ast_copy_string(line->protocol, rowval[i], sizeof(line->protocol));
					} else if (!strcmp(fields[i].name, "line_prefix")) {
						ast_copy_string(line->prefix, rowval[i], sizeof(line->prefix));
					} else if (!strcmp(fields[i].name, "line_postfix")) {
						ast_copy_string(line->postfix, rowval[i], sizeof(line->postfix));
					}
				}
			}
			memset(dialargs, 0, sizeof(dialargs));
			if (!build_dial_string(line, args.number, dialargs, sizeof(dialargs))) {
				char *cursor = dialargs + strlen(dialargs);
				size_t left = sizeof(dialargs) - strlen(dialargs);
				if (!ast_strlen_zero(args.dialopts) && (left > 2)) {
					ast_build_string(&cursor, &left, AST_STANDARD_APP_DELIM);
					ast_build_string(&cursor, &left, args.dialopts);
				}
				if (debug)
					ast_log(LOG_NOTICE, "Dialing: %s\n", dialargs);
				res = pbx_exec(chan, appDial, dialargs);

				if (res == 0) {	//Dial application returned without hangup
					dialstatus = pbx_builtin_getvar_helper(chan, "DIALSTATUS");
					if (debug) {
						ast_log(LOG_NOTICE, "Dial returned with status: %s\n",
								dialstatus);
					}
					if (!strcmp(dialstatus, "BUSY")) {	// Dial returns BUSY on this line
						if (failonbusy) {
							ast_indicate(chan, AST_CONTROL_BUSY);
							/* give em busy tone for several seconds */
							time(&rstart);
							while (time(NULL) < rstart + 10) {
								if (ast_waitfor(chan, -1) < 0)
									break;
								frm = ast_read(chan);
								if (!frm)
									break;
								ast_frfree(frm);
							}
							ast_playtones_stop(chan);
							goto out;
						} else {
							ast_log(LOG_NOTICE,
									"Dial returned BUSY on line %s. Going to next line...\n",
									line->number);
						}
					} else if (!strcmp(dialstatus, "NOANSWER")) {	// Dial Not Answered
						ready = 1;
						break;
					}
				} else {
					ast_log(LOG_DEBUG, "Dial result: %d\n", res);
					ready = 1;
					break;
				}
			}
		}						// Fin iteracion sobre HUNTGROUP
	}							// Fin if(res == 0 && numlines > 0)
	if (!ready) {				// Dial not completed but we have no more lines
		ast_log(LOG_NOTICE, "No more lines... DialHuntGroup returned NOMORELINES\n");
		pbx_builtin_setvar_helper(chan, "DIALHUNTGROUPSTATUS", "NOMORELINES");
	}
  out:
	if (sqlres) {
		mysql_free_result(sqlres);
		sqlres = NULL;
	}
	if (line)
		free(line);
	ast_module_user_remove(u);
	return res;
}

static int vpbxgoto_exec(struct ast_channel *chan, void *data)
{
	int res = 0;
	struct virtual_pbx *p = NULL;
	char target[256];
	int vpbx = 0;
	char destflag;
	struct ast_module_user *u;

	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "VirtualPBXGoto requires a flag argument)\n");
		return -1;
	}

	u = ast_module_user_add(chan);

	vpbx = find_current_vpbx(chan);
	p = get_virtual_pbx(vpbx);
	if (!p) {
		ast_log(LOG_ERROR, "Virtual PBX %d not configured!", vpbx);
		res = -1;
		goto out;
	}

	destflag = ((char *) data)[0];
	switch (destflag) {
	case 'm':
		ast_verbose(VERBOSE_PREFIX_3 "Virtual PBX Goto:  mainmenu\n");
		res = ast_parseable_goto(chan, (char *) p->mainmenu);
		break;
	case 'o':
		ast_verbose(VERBOSE_PREFIX_3 "Virtual PBX Goto:  operator\n");
		res = ast_parseable_goto(chan, (char *) p->operator);
		break;
	case 'v':
		ast_verbose(VERBOSE_PREFIX_3 "Virtual PBX Goto:  voicemail\n");
		if (IS_INCOMING_CALL(chan))
			snprintf(target, sizeof(target), "%s" AST_STANDARD_APP_DELIM "%s", (char *) p->vmsincoming, "s" AST_STANDARD_APP_DELIM "1");
		else
			snprintf(target, sizeof(target), "%s" AST_STANDARD_APP_DELIM "%s", (char *) p->vmslocal, "s" AST_STANDARD_APP_DELIM "1");
		res = ast_parseable_goto(chan, target);
		break;
	default:
		ast_log(LOG_WARNING, "VirtualPBXGoto argument must be 'm', 'o' or 'v')\n");
		res = -1;
		break;
	}
  out:
	ast_module_user_remove(u);
	return res;
}

inline int build_dial_string(struct line_info *line, const char *number, char *dialstr,const int len)
{
	int ret = 0;
	if (!strcasecmp(line->protocol, "zap") || !strcasecmp(line->protocol, "dahdi")) {
		snprintf(dialstr, len, "%s/%s/%s", line->protocol, line->prefix, number);
	} else if (!strcasecmp(line->protocol, "sip")) {
		snprintf(dialstr, len, "sip/%s%s%s", line->prefix, number, line->postfix);
	} else if (!strcasecmp(line->protocol, "h323")) {
		snprintf(dialstr, len, "h323/%s%s", line->prefix, number);
	} else if (!strcasecmp(line->protocol, "oh323")) {
		snprintf(dialstr, len, "oh323/%s%s", line->prefix, number);
	} else if (!strcasecmp(line->protocol, "capi")) {
		if (!ast_strlen_zero(line->postfix)) {
			snprintf(dialstr, len, "capi/%s/%s/%s", line->prefix, number, line->postfix);
		} else {
			snprintf(dialstr, len, "capi/%s/%s", line->prefix, number);
		}
	} else if (!strcasecmp(line->protocol, "vpb")) {
		snprintf(dialstr, len, "vpb/%s/%s", line->prefix, number);
	} else {
		ast_log(LOG_WARNING, "Unknown protocol %s on line %s\n", line->protocol,
				line->number);
		ret = -1;
	}
	if (!ret && debug) {
		ast_log(LOG_NOTICE, "Line: protocol => %s prefix => %s postfix => %s\n",
				line->protocol, line->prefix, line->postfix);
		ast_log(LOG_NOTICE, "Channel Dial String: %s\n", dialstr);
	}
	return ret;
}


/* load_config:
   Load configuration from dialipcontact.conf 
   Returns 0 on success or -1 when an error is found.\n"
*/
static int load_config(void)
{
	struct ast_config *config;
	struct ast_variable *var;
	char config_path[PATH_MAX] = "";

	snprintf(config_path, sizeof(config_path), "%s/%s", IPCONTACT_CONFDIR, config_file);
	config_path[sizeof(config_path) - 1] = 0;

  /***** BEGIN MUTEX BLOCK **********/
	ipc_mutex_lock();
#ifdef HAS_AST16
	struct ast_flags config_flags = { CONFIG_FLAG_WITHCOMMENTS };
	config = ast_config_load((char *) config_path, config_flags);
#else
	config = ast_config_load((char *) config_path);
#endif
	if (!config) {
		ast_log(LOG_WARNING, "Unable to load config file: %s\n", config_path);
		return AST_MODULE_LOAD_DECLINE;
	};


	memset(&callinfo, 0, sizeof(struct callinfo_config));
	memset(&dialhunt, 0, sizeof(struct dialhunt_config));

	if (ast_true(ast_variable_retrieve(config, "general", "debug"))) {
		ast_log(LOG_NOTICE, "%s loaded in debug mode\n", MODULE_DESC);
		debug = 1;
	} else {
		debug = 0;
	}

	// Parsing callinfo section
	var = ast_variable_browse(config, "callinfo");
	while (var) {
		if (!strcasecmp(var->name, "usedb")) {
			callinfo.usedb = ast_true(var->value);
		}
		if (!strcasecmp(var->name, "usemanager")) {
			callinfo.usemanager = ast_true(var->value);
		}
		var = var->next;
	}

	// Parsing dialhunt section
	var = ast_variable_browse(config, "dialhuntgroup");
	while (var) {
		if (!strcasecmp(var->name, "failonbusy")) {
			dialhunt.failonbusy = ast_true(var->value);
		}
		if (!strcasecmp(var->name, "dialopts")) {
			ast_copy_string(dialhunt.dialopts, var->value, sizeof(dialhunt.dialopts));
		}
		var = var->next;
	}

	ast_config_destroy(config);
	ipc_mutex_unlock();
  /***** END MUTEX BLOCK **********/

	return 0;
}

/*
 * Standard module functions ...
 */
static int load_module(void)
{
	int res = 0;
	res |=
		ast_register_application(app_dialhunt, dialhunt_exec, synopsis_dialhunt,
								 descrip_dialhunt);
	res |=
		ast_register_application(app_savecallinfo, savecallinfo_exec,
								 synopsis_savecallinfo, descrip_savecallinfo);
	res |=
		ast_register_application(app_vpbxgoto, vpbxgoto_exec, synopsis_vpbxgoto,
								 descrip_vpbxgoto);
	res |=
		ast_register_application(app_callforward, callforward_exec, synopsis_callforward,
								 descrip_callforward);
	res |=
		ast_register_application(app_setcidext, setcidext_exec, synopsis_setcidext,
								 descrip_setcidext);
	res |=
		ast_register_application(app_get_pin_profile, get_pin_profile_exec,
								 synopsis_get_pin_profile, descrip_get_pin_profile);
	res |=
		ast_register_application(app_get_exten_profile, get_exten_profile_exec,
								 synopsis_get_exten_profile, descrip_get_exten_profile);
	if (res == 0) {
		if (!(appDial = pbx_findapp("Dial"))) {
			res = AST_MODULE_LOAD_DECLINE;
		} else {
			res = load_config();
		}
	}
	return res;
}

static int unload_module(void)
{
	int res = 0;
	res |= ast_unregister_application(app_dialhunt);
	res |= ast_unregister_application(app_setcidext);
	res |= ast_unregister_application(app_savecallinfo);
	res |= ast_unregister_application(app_vpbxgoto);
	res |= ast_unregister_application(app_callforward);
	res |= ast_unregister_application(app_get_pin_profile);
	res |= ast_unregister_application(app_get_exten_profile);
	return res;
};


AST_MODULE_INFO_STANDARD(ASTERISK_GPL_KEY, MODULE_DESC);
